#if 0

  Configuration of pieces for any orientation

#endif

#include <memory.h>
#include "stdefun.h"
#include "orient.h"
#include "orients.h"
#include "config.h"
#include "debug.h"


/* vec3 = mat3x3 * vec3 */

static void
qr_mult_ixyz  _P3(
   (T3D_IXYZ *,     o3),
   (T3D_QUAD_ROT *, qr),
   (T3D_IXYZ *,     i3))
{
   int  i, j, v;

   for (i = 0;  i < 3;  i++) {
      j = qr->oxyz[i].axe;
      v = i3->v[i];
      /*  Assuming v>=0, v represents [v,v+1], 
       *  Xfromed to [-(v+1),v], represented by  (v-1).
       */
      DASSERT(v >= 0);
      o3->v[ j ] = (qr->oxyz[i].dir == 0  ? v  : -(v + 1));
   }
} /* qr_mult_ixyz */


/* piece = mat3x3 * piece */
static void
orient_piece  _P3(
   (T3D_4CUBES*,    opiece),
   (T3D_QUAD_ROT *, qr),
   (T3D_4CUBES*,    ipiece))
{
   int  ci;

   for (ci = 0;   ci < 4;  ci++) {
      qr_mult_ixyz( &opiece->cube[ci], qr, &ipiece->cube[ci]);
   }
} /* orient_piece */


/* Xlate piece so it will be in the positive octant, centered in [0,4)^3 */
static void
positive_center  _P2(
   (T3D_4CUBES *, piece),
   (T3D_IXYZ *,   shift))
{
   int  xi, ci, m, M, v, t;

   for (xi = 0;  xi < 3;  xi++) {

      /* find minimum & Maximum of xi'th axe of all 4 cubes */
      m = M = piece->cube[0].v[xi];
      for (ci = 3;  ci != 0;  ci--) {
         v = piece->cube[ci].v[xi];
         if (v < m) {
            m = v; 
         } else if (M < v) {
            M = v;
         }
      }
      DASSERT( m <= M  &&  M - m < 4);
            
      /* to make the median (m+M)/2 (lowerly rounded) to move into 1 */
      t = 1 - (((m + M + 8) / 2) - 4);

      for (ci = 0;  ci < 4;  ci++) { /* Xlate */
         piece->cube[ci].v[xi] += t;
      }
      shift->v[xi] = t;
   }
} /* positive_center */


static void
mask_config  _P2(
   (T3D_CONFIG *, config_mask),
   (T3D_4CUBES *, piece))
{
   int          ci, zi;
   
   /* Zero masks */
   for (zi = 0;  zi < 4;  zi++) {
      config_mask->mask_level[zi] = 0;
   }

   /* Add mask bits */
   for (ci = 0;  ci < 4;  ci++) {
      T3D_IXYZ  *c = &piece->cube[ci];
      config_mask->mask_level[ c->v[2] ] |= T3D_XY_BIT( c->v[0], c->v[1]);
   }

   /* Compute Bounding box */
   for (zi = 0;  zi < 3;  zi++) {
      int  m, M, v;
      m = M = piece->cube[3].v[zi];
      ci = 3;
      do {
         ci--;
         v = piece->cube[ci].v[zi];
         if (v < m) {
            m = v;
         } else if (M < v) {
            M = v;
         }
      } while (ci != 0);
      config_mask->minmax[0].v[zi] = m;
      config_mask->minmax[1].v[zi] = M;
   }
} /* mask_config */


void
t3d_make_piece_config  _P2(
   (T3D_CONFIG_ORIENT *, co),
   (T3D_4CUBES *,        piece))
{
   int         oi;
   T3D_MATQR    *orient;

#if DEBUG
{T3D_4CUBES   opiece;
 t3d_get_orientation(&orient, 0);
 orient_piece( &opiece, &orient->qr, piece); /* Identity */
 positive_center( &opiece, &co->orient[0].shift);
 if (memcmp(&opiece, piece, sizeof(T3D_4CUBES)) != 0) {
    fprintf(stderr, "Original piece is not positively centered!\n");
 }}
#endif /* DEBUG */

   
   for (oi = 0;  oi < T3D_N_ORIENT;  oi++) {
      T3D_4CUBES  *p4c = &co->orient[oi].c;
      t3d_get_orientation(&orient, oi);
      orient_piece( p4c, &orient->qr, piece);
      positive_center( p4c, &co->orient[oi].shift ); 
      mask_config( &co->orient[oi] , p4c);
   }
} /* t3d_make_piece_config */
