#include <math.h>
#include <X11/Intrinsic.h>   /* Just for:   XtOffset  !!! */
#include <phigs/phigs.h>
#include "stdefun.h"
#include "baseid.h"
#include "colorid.h"
#include "orients.h"
#include "space.h"
#include "debug.h"

typedef struct {
   int  i[3][2];
} INT3_2;


static void 
t3d_cube  _P3(
   (int,      major_axe),
   (int,      major_side),
   (INT3_2 *, color6)
)
{
   typedef Ppoint3    *pPpoint3;
   static float       zo[] = {0., 1.};
   static int         fadd[] = {
                         XtOffset(pPpoint3, x),
                         XtOffset(pPpoint3, y),
                         XtOffset(pPpoint3, z)
                      };
   static int         ow[2][4] = {  /* order to define outer normal */
                         {0, 1, 2, 3},
                         {3, 2, 1, 0}
                      };
   int                xi,        /* axe of face */ 
                      l,         /* 0,1 - two level faces per axe */
                      /* fi,        /* 3*2 = 6 faces */
                      xi0, xi1,  /* The other varying axes of the face */
                      oxi,   	 /* orientation index */
                      i,
                      hsr_on = t3d_hsr_on();
   Ppoint3            p[4];
   Ppoint_list3       pl;  
   Ppoint_list_list3  pll;


#define PA(p,xi) ((Pfloat *)(((char *)(&p)) + fadd[xi]))


   pl.num_points = 4;
   pl.points = p;
   pll.num_point_lists = 1;
   pll.point_lists = &pl;
#if 0
   pset_int_style( PSTYLE_SOLID );
   pset_face_cull_mode( PCULL_NONE );
   pset_face_disting_mode( PDISTING_NO);
#endif

   /*  major_axe = the major axis, for which we will draw 
    *  only one-front face major_side to be the last one,
    *  thus achieving ~HSR if not available by the implementation.
    */
   for (oxi = 2;  oxi >= 0;  oxi--) {
      /* For the last oxi, oxi == 0, Have  xi = oi / 2 */
      xi = (major_axe + oxi) % 3;
      xi0 = (xi + 1) % 3;
      xi1 = (xi0 + 1) % 3;
      /* Next l-loop, goes 0,1, but for the last oxi==0, only l = major_side */
      for ( l =  (oxi != 0 ? 0 : (hsr_on ? 0 : major_side));
            l <= (oxi != 0 ? 1 : (hsr_on ? 1 : major_side));
            l++) {
         /* fi = 2 * xi + oxi;   DASSERT(0 <= fi && fi < 6); */
         pset_int_colr_ind(color6->i[xi][l]);
         for (i = 0;  i < 4;  i++) {
            *PA(p[i], xi) = zo[l];
         }

         *PA(p[ ow[l][0] ], xi0) = 0.;  *PA(p[ ow[l][0] ], xi1) = 0.; 
         *PA(p[ ow[l][1] ], xi0) = 1.;  *PA(p[ ow[l][1] ], xi1) = 0.;  
         *PA(p[ ow[l][2] ], xi0) = 1.;  *PA(p[ ow[l][2] ], xi1) = 1.;  
         *PA(p[ ow[l][3] ], xi0) = 0.;  *PA(p[ ow[l][3] ], xi1) = 1.;  

#        ifdef DEBUG0
          DHERE; printf("fi=%d clr=%d =>", (2*xi + l), color6->i[xi][l]);
          for (i = 0;  i < 4;  i++) {
             printf(" [%g %g %g] ", p[i].x, p[i].y, p[i].z);
          }
          printf("\n");
#        endif

         pfill_area_set3( &pll );
      }
   }
   DPR0((stderr, "\n"));
} /* t3d_cube */


   
void
t3d_cube_orient  _P3(
   (double, x),
   (double, y),
   (double, z)
)
{
   static int curr_vmajor_axe = -1, curr_vmajor_side = -1;
   /* Ugly, but consistent color */
   static INT3_2  face_color = { { {T3D_CLR_RED,  T3D_CLR_GREEN}, 
                                   {T3D_CLR_BLUE, T3D_CLR_YELLOW}, 
                                   {T3D_CLR_CYAN, T3D_CLR_MAGENTA} 
                               } };
                      
   int     i, j, oi, vmajor_axe = -1, vmajor_side;
   INT3_2  color;
   double  max, xyz[3], abs;

   DASSERT((curr_vmajor_axe == -1) == (curr_vmajor_side == -1));
   if (curr_vmajor_axe != -1 && t3d_hsr_on()) { /* if HSR no work needed */
      return;
   }
      
   xyz[0] = x;
   xyz[1] = y;
   xyz[2] = z;


   for (i = 0, max = -1.;  i < 3;  i++) {
      abs = fabs( xyz[i]);
      if (max < abs) {
         max = abs;
         vmajor_axe = i;
      }
   }
   DASSERT(0 <= vmajor_axe && vmajor_axe < 3);
   vmajor_side =  (xyz[vmajor_axe] > 0 ? 1 : 0);
   DPR0((stderr, "N[vmajor_axe=%d]=%g\n", vmajor_axe, xyz[vmajor_axe]));

   if (curr_vmajor_axe != vmajor_axe || curr_vmajor_side != vmajor_side) {
      curr_vmajor_axe = vmajor_axe;
      curr_vmajor_side = vmajor_side;

      for (oi = 0;  oi < T3D_N_ORIENT;  oi++) {
         T3D_MATQR  *matqr;
         t3d_get_orientation( &matqr, oi);

         /* permutate the colors according to the Xformation,
          * and see who get to the major-axe.
          */
         for (i = 0;  i < 3;  i++) {
            int  xi = matqr->qr.oxyz[i].axe,
                 si = matqr->qr.oxyz[i].dir;
            for (j = 0;  j < 2;  j++) {
               color.i[xi][j] = t3d_colorid( face_color.i[i][ (si != j) ] );
            }
         }

         pset_edit_mode( PEDIT_INSERT);
         DPR0((stdout, "\nT3D_ID_ORIENT_CUBE+%d\n", oi));
         popen_struct( T3D_ID_ORIENT_CUBE + oi );
            pset_int_style( PSTYLE_SOLID);
            pset_edge_flag( PEDGE_ON);
            pset_edge_colr_ind( 1 ); /* White */
            t3d_cube( vmajor_axe, vmajor_side, &color);
         pclose_struct();
      }
      DCODE0(ppost_struct(1, T3D_ID_ORIENT_CUBE, 1.);)
      pupd_ws( 1, PFLAG_PERFORM);
   }
} /* t3d_cube_orient */
      
