#include <stddef.h>
#include "stdefun.h"
#include "assign.h"
#include "baseid.h"
#include "config.h"
#include "fourcube.h"
#include "game.h"
#include "instance.h"
#include "orient.h"
#include "orients.h"
#include "pieces.h"
#include "rotquad.h"
#include "score.h"
#include "debug.h"


typedef struct {
   int      type,
            orient;
   T3D_IXYZ  move;
} T3D_INSTANCE;


static int  game_on = 0;      /* piece(s) exist but not full space */
static int  game_active = 0;  /* not paused! */
static int  curr_piece = -1;

static short  space_ocubes[15][4][4];  /* space_ocubes[z][y][x] holds
                                        * the orientation of a cube in [x,y,z],
                                        * -1 if not exist!
                                        */
static T3D_LEVEL_MASK  game_zlevel[15]; /* Compact existance mask of the above */

static T3D_INSTANCE  curr_instance = {-1, -1, {-1, -1, -1}};

#if 0
  xy_side_width [0], [1], are two sets of mask for x,y respectively.
  Each to be considered indexed with [-3..+3], so logical-index 0 is maped to 3.
  Logical index of +n (-n) is a mask of the n high (low) sides for x or y.
#endif
#define M3(i)  ((i)+3)
static T3D_LEVEL_MASK  xy_side_width[2 /* XY */][ 3 - (-3) + 1];


/* Scoring factors */
static int  n_pieces, n_levels;


static void (*fp_endgame_callback)  ARGS((void)) = NULL;



#ifdef ASSERT
space_zlevel_consistent()
{
   T3D_LEVEL_MASK  mask;
   int  xi, yi, zi;

   for (zi = 0;  zi < 15;  zi++) {
      mask = 0;
      for (yi = 0;  yi < 4;  yi++) {
         for (xi = 0;  xi < 4;  xi++) {
           if (space_ocubes [zi] [yi][xi] != -1) {
              mask |= T3D_XY_BIT( xi, yi);
           }
         }
      }
      if (mask != game_zlevel[zi]) { DHERE;
         fprintf(stderr, "zi=%d, soc=%ld != gzl=%ld\n",
                 zi, mask, game_zlevel[zi]);  DASSERT(0);
      }
   }
} /* space_zlevel_consistent */
#  define SPACE_ZLEVEL_CONSISTENT space_zlevel_consistent()
#else
#  define SPACE_ZLEVEL_CONSISTENT
#endif

static void
reduce  _P2(
   (int,  z0),
   (int,  zb)
)
{
   static T3D_LEVEL_MASK full_level = 0; /* constant to be initialized */
   int  z, zi, zd,
        nr,   /* current # reduces */
        tr,   /* total # reduces */
        x, y,
        oi;

   t3d_instance_settle(1);
   if (full_level == 0) {
      for (x = 0;  x < 4;  x++) {
         for (y = 0;  y < 4;  y++) {
            full_level |= T3D_XY_BIT( x, y);   
         }
      }
      DPR0((stderr, "full_level = 0x%x\n", full_level));
   }

   for (z = z0, tr = 0;  z + tr < zb; /* increment inside */) {
      /* Packed drop. Think of a case of a Long 4-cubes, whose 1st, 2nd, 4th
       * fill Z-levels, but not the 3rd. We want two drops for such case:
       * i.e.  nr (upon exit the following for-loop) = 2, 1.
       */
      for (nr = 0;  (zi = z + nr) < zb && game_zlevel[zi] == full_level; nr++) {
         DPR0((stderr, "Clearing Zi=%d\n", zi));
         for (x = 0;  x < 4;  x++) {
            for (y = 0;  y < 4;  y++) {
               DASSERT( space_ocubes [zi] [y][x] >= 0);
            }
         }
      }

      if (nr > 0) {

         /* drop instances, do it separately to save t3d_instance() calls! */
         for (x = 0;  x < 4;  x++) {
            for (y = 0;  y < 4;  y++) {
               for (zi = z;  (zd = zi + nr) < 15;  zi++) {
                   /* drop level zd into zi */
                  oi = space_ocubes [zd] [y][x];
                  if (space_ocubes [zi] [y][x] != oi) {
                     if (space_ocubes [zi] [y][x] >= 0) {
                        /* Could avoid this, once I know to do Phigs/replace */
                        t3d_instance(x, y, zi, -1);
                     }
                     t3d_instance(x, y, zi, oi);
                     space_ocubes [zi] [y][x] = oi;
                  }
               }
               /* clear top */
               for (; zd < 15;  zd++) {
                  if (space_ocubes [zd] [y][x] != -1) {
                     t3d_instance(x, y, zd, -1);
                     space_ocubes [zd] [y][x] = -1;
                  }
               }
            }
         }

         /* drop game_zlevel */
         for (zi = z;  (zd = zi + nr) < 15;  zi++) {
            game_zlevel[zi] = game_zlevel[zd];
         }
         for (; zi < 15;  zi++) {
            game_zlevel[zi] = 0;
         }

         tr += nr;  /* increment */
      } else {
         z++;       /* increment */
      }  
   }
   if (tr > 0) {
      t3d_score_add_event_value( T3D_SCOREVENT_1LEVEL - 1 + tr, n_levels);
      n_levels++;
      t3d_instance_flush();
   }
   t3d_instance_settle(0);
} /* reduce */


static void
do_instance  _P2(
   (T3D_INSTANCE *, instance),
   (int,           on)
) 
{
   T3D_CONFIG  *pconfig;
   int        ci, i, oi;
   T3D_IXYZ    loc;

   pconfig = &(t3d_piece_config( instance->type )->orient[ instance->orient ]);
   oi = (on ? instance-> orient : -1);
   for (ci = 0;  ci < 4;  ci++) {
      for (i = 0;  i < 3;  i++) {
         loc.v[i] = pconfig->c.cube[ci].v[i] + instance->move.v[i];
      }
      t3d_instance( loc.v[0],  loc.v[1],  loc.v[2], oi);
      space_ocubes [ loc.v[2] ] [ loc.v[1] ][ loc.v[0] ] = oi;
   }
} /* t3d_new_piece */


static void
settle  _P1( 
   (T3D_INSTANCE *, pinstance)
)
{
   int               zi, mzi, mz0, xyshift;
   T3D_CONFIG         *pcm;
   T3D_LEVEL_MASK     add;

   do_instance( pinstance, 0);
   t3d_instance_settle( 1);
   do_instance( pinstance, 1);
   t3d_instance_settle( 0);

   pcm = &(t3d_piece_config( pinstance->type )->orient[ pinstance->orient ]);

   xyshift = T3D_XY_POS(pinstance->move.v[0], pinstance->move.v[1]);
   for (zi = pcm->minmax[0].v[2], mzi = mz0 = pinstance->move.v[2] + zi;  
        zi <= pcm->minmax[1].v[2];  
        zi++, mzi++) {
      DASSERT(0 <= mzi && mzi < 15);
      add = pcm->mask_level[zi];
      DASSERT(add != 0);
      add = (xyshift >= 0 ?  add << xyshift  :  add >> (-xyshift));
      game_zlevel[ mzi ] |= add;
   }
   SPACE_ZLEVEL_CONSISTENT;
   reduce(mz0, mzi);
   SPACE_ZLEVEL_CONSISTENT;
#if DEBUG0
{int x, y;
  printf("zlevel[0]=%x\n", game_zlevel[0]);
  for (y=3; y>=0; y--) {
   for (x=0; x<4; x++) {
    putchar( (space_ocubes[0][y][x] >= 0 ? '*' : ' ') );
   } printf("\n");
  }}
#endif

} /* settle */



static void
clear_instances()
{
   static int virgin = 1;

   int  xi, yi, zi;

   if (virgin) {
      virgin = 0;
      for (zi = 0;  zi < 15;  zi++) {
         for (yi = 0;  yi < 4;  yi++) {
            for (xi = 0;  xi < 4;  xi++) {
               space_ocubes [zi] [yi][xi] = -1;
            }
         }
      }
   } else {
      if (game_on) {
         settle( &curr_instance);
      }
      SPACE_ZLEVEL_CONSISTENT;
      for (zi = 0;  zi < 15;  zi++) {
        if (game_zlevel[zi] != 0) {
           for (yi = 0;  yi < 4;  yi++) {
              for (xi = 0;  xi < 4;  xi++) {
                 if (space_ocubes [zi] [yi][xi] != -1) {
                    t3d_instance( xi, yi, zi, -1);
                    space_ocubes [zi] [yi][xi] = -1;
                 }
              }
           }
        }
     }
     t3d_instance_flush();
  }
} /* clear_instances */


static void
init_xy_side_width()
{
   int            xy, xyi, n;
   T3D_LEVEL_MASK  mask4[4], mask;


   for (xy = 0;  xy < 2;  xy++) {
      for (xyi = 0, mask = 0;  xyi < 4;  xyi++) {
         for (n = 0, mask = 0;  n < 4;  n++) {
            T3D_LEVEL_MASK  bit = (xy == 0 ?  T3D_XY_BIT(xyi, n)
                                          :  T3D_XY_BIT(n, xyi) );
            mask |= bit;
         }
         mask4[xyi] = mask;
      }
      for (n = -3;  n < 4;  n++) {
         int  low, high;
         if (n >= 0) {
            high = 3;
            low = 4 - n;
         } else {
            low = 0;
            high = -1 - n;
         }
         /*  0 => {},  1 => {3},  2 => {2,3},  3 => {1,2,3}
          *           -1 => {0}, -2 => {0,1}, -3 => {0,1,2}
          */
         for (xyi = low, mask = 0; xyi <= high;  xyi++)  mask |= mask4[xyi];
         xy_side_width[xy][M3(n)] = mask;
      }
   }

#if DEBUG
   for (xy = 0;  xy < 2;  xy++) {
      DASSERT(xy_side_width[xy][M3(0)] == 0);
      for (n = 1;  n < 4;  n++) {
         T3D_LEVEL_MASK a, o;
         a = ( xy_side_width[xy][M3(n)] & xy_side_width[xy][M3(-(4-n))]);
         DASSERT( a == 0);
         o = ( xy_side_width[xy][M3(n)] | xy_side_width[xy][M3(-(4-n))]);
         DASSERT( o == (((T3D_LEVEL_MASK)1L << 16) -1));
      }
   }
#endif /* DEBUG */
} /* init_xy_side_width */


/* Return if an oriented configurartion with a shift
 * intersects with oither game instances, or outside the boundary.
 */
static int
intersection  _P2(
   (T3D_CONFIG *, pcm),
   (T3D_IXYZ *,   move))
{
   int            zi, mzi, intersect, movex, movey, xyshift;
   T3D_LEVEL_MASK  conf_zmask;

   movex = move->v[0];
   movey = move->v[1];
   intersect = !(-4 < movex && movex < 4 && -4 < movey && movey < 4);
   for (zi = 0;  zi < 4 && !intersect;  zi++) {
      conf_zmask = pcm->mask_level[zi];
      if (conf_zmask != 0) {
         mzi = move->v[2] + zi;
         intersect = (mzi < 0 || mzi >= 15);  /* inside Z-range */
         if (!intersect) {
            intersect =
               /* inside XY-range */
               ((conf_zmask & xy_side_width[0][M3(movex)]) != 0) ||
               ((conf_zmask & xy_side_width[1][M3(movey)]) != 0);
               /* not colliding with others */
            if (!intersect) {
               xyshift = T3D_XY_POS(movex, movey);
               conf_zmask = (xyshift >= 0 ? (conf_zmask << xyshift)
                                          : (conf_zmask >> (-xyshift)) );
               intersect = ((conf_zmask & game_zlevel[mzi]) != 0);
            }
         }
      }
   }
   return(intersect);
} /* intersection */


static int 
rand_mod  _P1(
   (int,  m)
)
{
   int  r = rand();
   r >>= 8;
   r &= 0xFFFF;
   r = r % m;
   return(r);
} /* rand_mod */



static void
t3d_new_piece()
{
   int               type_id, orient_id;
   T3D_CONFIG_ORIENT  *pco;

   type_id = rand_mod(T3D_PIECE_N);
   orient_id = rand_mod(T3D_N_ORIENT);
   DCODE0({type_id = T3D_PIECE_LONG - T3D_PIECE_FIRST; 
                  orient_id = 0;}); /* Force long in Identity */
   DCODE0({type_id = T3D_PIECE_ELL-T3D_PIECE_FIRST; 
                   orient_id = 12;}); /* Force Ell in Identity */
   DCODE0({type_id = T3D_PIECE_BLOCK-T3D_PIECE_FIRST; 
                   orient_id = 0;}); /* Force Block in Identity */
   DCODE0({type_id = 0; orient_id = 8;}); /* Force long in z-dir */
   type_id += T3D_PIECE_FIRST;
   curr_instance.type = type_id;
   curr_instance.orient = orient_id;

   /* move up to be alligned to the top */
   curr_instance.move.v[0] =
   curr_instance.move.v[1] = 0;
   curr_instance.move.v[2] =
      14 - t3d_piece_config( type_id )->orient[orient_id].minmax[1].v[2];

   pco = t3d_piece_config(type_id);
   if (intersection( &pco->orient[orient_id], &curr_instance.move)) {
      DHERE0;  DPR0((stderr, "Game Over!\n"));
      game_on = 0;
      DASSERT( fp_endgame_callback != NULL);
      (*fp_endgame_callback)();
      t3d_score_record();
   } else {
      do_instance( &curr_instance, 1);
      t3d_instance_flush();
      t3d_score_add_event_value( T3D_SCOREVENT_NEW_PIECE, n_pieces);
      n_pieces++;
   }
} /* t3d_new_piece */


void
t3d_new_game()
{
   static int virgin = 1;
   int        i;

   if (virgin) {
      virgin = 0;
      t3d_genpieces();
      init_xy_side_width();
   }

   clear_instances();
   game_on = 1;
   for (i = 0;  i < 15;  i++) {
      game_zlevel[i] = 0;
   }
   n_pieces = n_levels = 0;
   t3d_score_set_value( 0);
   t3d_new_piece();
} /* t3d_new_game */


void 
t3d_game_set_active  _P1(
   (int, f)
)
{
   game_active = f;
} /* t3d_game_set_active */


void
t3d_fall()
{
   T3D_INSTANCE       *pinstance;
   T3D_CONFIG_ORIENT  *pco;

   DHERE0;
   if (game_on && game_active) {
      pinstance = &curr_instance;
      pco = t3d_piece_config( pinstance->type );

      pinstance->move.v[2] -= 1;  /* may it fall..? */
      if (intersection( &pco->orient[ pinstance->orient ], &pinstance->move)) {
         DHERE0;  DPR0((stderr, "Cannot fall anymore!\n"));
         pinstance->move.v[2] += 1;  /* restore */
         settle( pinstance );
         t3d_new_piece();
      } else { /* fall */
         pinstance->move.v[2] += 1;  /* restore for deletion */
         do_instance(pinstance, 0);  /* delete */
         pinstance->move.v[2] -= 1;  /* new fallen */
         do_instance(pinstance, 1);
         t3d_instance_flush();
      }
   }
} /* t3d_fall */


void
t3d_xymove  _P1(
   (T3D_MV_ENUM,  imove)
)
{
   static  T3D_IXYZ   move[] = { /* indexed by T3D_MV_ENUM */
                        {  1.,  0.,  0.},
                        { -1.,  0.,  0.},
                        {  0.,  1.,  0.},
                        {  0., -1.,  0.}
                     };
   T3D_IXYZ           save, movesave;
   T3D_INSTANCE       *pinstance;
   T3D_CONFIG_ORIENT  *pco;


   DASSERT( (sizeof(move) / sizeof(move[0]) == T3D_MV_SIZE) );
   DASSERT(0 <= imove && imove < T3D_MV_SIZE);
   DHERE0;
   if (game_on && game_active) {
      pinstance = &curr_instance;
      pco = t3d_piece_config( pinstance->type );

      save.v[0] = pinstance->move.v[0];
      save.v[1] = pinstance->move.v[1];

      movesave.v[0] = pinstance->move.v[0] += move[imove].v[0];
      movesave.v[1] = pinstance->move.v[1] += move[imove].v[1];
      /*                 [2] == Z - does not move */

      if (intersection( &pco->orient[ pinstance->orient ], &pinstance->move)) {
         DPR0((stderr, "Cannot move anymore!\n"));
         pinstance->move.v[0] = save.v[0];
         pinstance->move.v[1] = save.v[1];
      } else { /* move */

         pinstance->move.v[0] = save.v[0];
         pinstance->move.v[1] = save.v[1];
         do_instance( pinstance, 0 );

         pinstance->move.v[0] = movesave.v[0];
         pinstance->move.v[1] = movesave.v[1];
         do_instance( pinstance, 1 );


         t3d_instance_flush();
      }
   }
} /* t3d_xymove */


void
t3d_axe_rotate  _P1(
   (T3D_ROTQUAD, rotaxe)
)
{
   T3D_IXYZ           old_move;
   T3D_INSTANCE       *pinstance, old_instance;
   T3D_CONFIG_ORIENT  *pco;
   int               old_oi, new_oi, upz, i, j;
   T3D_CONFIG         *pcm;


   DASSERT(0 <= rotaxe && rotaxe < T3D_N_ROTQUAD);
   DHERE0;
   if (game_on && game_active) {
      pinstance = &curr_instance;

      ASSIGN( &old_instance, pinstance);

      pco = t3d_piece_config( pinstance->type );

      /* save for possible restore */
      ASSIGN(&old_move, &pinstance->move);
      old_oi = pinstance->orient;

      pinstance->orient = new_oi = t3d_get_rotated_orient_id( old_oi, rotaxe);

      /*  Ideally, we would like to change only the orientation.
       *  But, we must ensure that it is inside the 4*4*15 box.
       */

      /* It might get upper than top */
      pcm = &pco->orient[new_oi];
      upz = old_move.v[2] + pcm->minmax[1].v[2];
      if (upz > 14) {
         pinstance->move.v[2] -= (upz - 14); /* lowering */
      }

      /* Now let it fit inside the XY boundary */
      for (i = 0;  i < 2;  i++) {
         int  v[2];
         for (j = 0;  j < 2;  j++) {
            v[j] = pcm->minmax[j].v[i] + pinstance->move.v[i];
         }
         /* v[0], v[[1], are the new min-max in the i-th axe. */
         DASSERT(v[0] <= v[1] && (v[1] - v[0]) < 4);
         if (v[0] < 0) {
            pinstance->move.v[i] += (0 - v[0]);  /* positive shift */
         } else if (3 < v[1]) {
            pinstance->move.v[i] -= (v[1] - 3);
         }
         DASSERT(0 <= pcm->minmax[0].v[i] + pinstance->move.v[i]);
         DASSERT(     pcm->minmax[1].v[i] + pinstance->move.v[i] < 4);
      }

      if (intersection( &pco->orient[ pinstance->orient ], &pinstance->move)) {
         DPR0((stderr, "Cannot rotate anymore!\n"));
         /* restore */
         ASSIGN( &pinstance->move, &old_move);
         pinstance->orient = old_oi;
      } else { /* rotate */
         do_instance( &old_instance, 0);
         do_instance( pinstance, 1);
         t3d_instance_flush();
      }
   }
} /* t3d_axe_rotate */


void
t3d_drop_down()
{
   T3D_INSTANCE       *pinstance;
   T3D_CONFIG_ORIENT  *pco;
   T3D_CONFIG         *pconfig;
   int               savez, dropz, ci, z, zi;

   DHERE0;
   if (game_on && game_active) {
      pinstance = &curr_instance;
      pco = t3d_piece_config( pinstance->type );
      savez = pinstance->move.v[2];
      pconfig = &pco->orient[ pinstance->orient ];

      /* find maximal z */
      z = pconfig->c.cube[0].v[2];
      ci = 3;
      do {
         zi = pconfig->c.cube[ci].v[2];
         if (z < zi) {
            z = zi;
         }
         ci--;
      } while (ci != 0);
      z += savez; /* pinstance->move.v[2] */
      t3d_score_add_event_value( T3D_SCOREVENT_DROP, z); 
      
      do {
         pinstance->move.v[2]--;
      } while (!intersection( &pco->orient[ pinstance->orient ],
                            &pinstance->move));
      pinstance->move.v[2]++;
      if (pinstance->move.v[2] != savez) { /* did drop */
         dropz = pinstance->move.v[2];  /* save drop */
         pinstance->move.v[2] = savez; 
         do_instance( pinstance, 0);    /* delete */
         pinstance->move.v[2] = dropz; 
         do_instance( pinstance, 1);
         t3d_instance_flush();
      }
      settle( pinstance );
      t3d_new_piece();
   }
} /* t3d_drop_down */


void
t3d_endgame_callback
#if __STDC__
                   (void (*f)())
#else
                   (f)
   void  (*f)();
#endif
{
   fp_endgame_callback = f;
} /* t3d_endgame_callback */


int
t3d_game_npiece()
{
   return( n_pieces);
} /* t3d_game_npiece */
