#include <phigs/phigs.h>
#include "stdefun.h"
#include "baseid.h"
#include "instance.h"
#include "orients.h"
#include "pieces.h"
#include "debug.h"


static int  settle = 0; /* flag for capping cubes */

static int
t3d_instance_index  _P3(
   (int, x),
   (int, y),
   (int, z)
)
{
   int  rc;

   DASSERT(0 <= x && x < 4);
   DASSERT(0 <= y && y < 4);
   DASSERT(0 <= z && z < 15);
   rc = (4*4) * z  +  4 * y  + x;
   return(rc);
} /* t3d_instance_index */


static void 
t3d_factorize_index _P4(
   (int *, px),
   (int *, py),
   (int *, pz),
   (int,   index)
) 
{
   *px = index % 4;
   *py = (index / 4) % 4;
   *pz = index / (4*4);
   DASSERT(index == t3d_instance_index( *px, *py, *pz));
} /* t3d_factorize_index */


/*  Magic formula for giving better values to higher z and xy-centered */
static double
priority_xyz  _P3(
   (int,  x),
   (int,  y),
   (int,  z)
) 
{
   int     p, dx, dy, dxy2;
   double  rc;

   dx = 2*x - 3;
   dy = 2*y - 3;
   dxy2 = dx*dx + dy*dy;
   DASSERT(0 <= dxy2 <= 18);
   p = 20 * z + (20 - dxy2);
   DASSERT(0 <= p && p < 300);
   rc = (1. / 300.) * (double)p;
   DPR0((stderr, "priority_xyz: %d -> %5.4f <-[%d, %d, %d]\n", p, rc, x, y, z));
   DASSERT(0. <= rc && rc < 1.);
   return( rc);
} /* priority_xyz */


void
t3d_instance_settle  _P1(
   (int, f)
)
{
   settle = f;
} /* t3d_instance_settle */



void
t3d_instance  _P4(
   (int, x),
   (int, y),
   (int, z),
   (int, orient)
) 
{
   static Pmatrix3      r = { {1., 0., 0., 0.},
                              {0., 1., 0., 0.},
                              {0., 0., 1., 0.},
                              {0., 0., 0., 1.} };
   int    index, sid;

   index = t3d_instance_index( x, y, z);
   DASSERT(0 <= index && index <= t3d_instance_index(3, 3, 14));
   DASSERT(-1 <= orient && orient < T3D_N_ORIENT);

   sid = T3D_ID_SPACE + index;
   DPR0((stderr, "oi=%d xyz=[%d %d %d], sid=%d\n", orient, x, y, z, sid));
   if (orient == -1) {
      punpost_struct( 1, sid);
      pdel_struct( sid);
   } else {
      r[0][3] = x;
      r[1][3] = y;
      r[2][3] = z; 
      DASSERT(0 <= x && x < 4);
      DASSERT(0 <= y && y < 4);
      DASSERT(0 <= z && z < 15);
      pset_edit_mode(PEDIT_INSERT);
      popen_struct( sid);
         pset_hlhsr_id( PHIGS_HLHSR_ID_ON);
         pset_view_ind( 1);
         pset_local_tran3( r, PTYPE_PRECONCAT);
         pexec_struct( T3D_ID_ORIENT_CUBE + orient);
         if (settle) {
            pexec_struct( T3D_ID_GRID_CAP + z);
         }
      pclose_struct();
      ppost_struct( 1, sid, priority_xyz( x, y, z) );
   }
} /* t3d_instance */


void
t3d_instance_flush()
{
   pupd_ws( 1, PFLAG_PERFORM);
} /* t3d_instance_flush */
