#include <stdio.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>

#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Simple.h>

#include "stdefun.h"
#include "action.h"
#include "game.h"
#include "scorei.h"
#include "space.h"
#include "timer.h"
#include "xcolorid.h"
#include "xview.h"
#include "debug.h"

typedef enum {
   PSRS_NONE,
   PSRS_PAUSE, 
   PSRS_RESTART  
} PSRS;

static Widget  pause_restart, 
               game;
static PSRS  psrs; /* pause_restart status */
static char  *XtNpause = "Pause",
             *XtNrestart = "Restart";

static void 
Destroyed  _P3( 
   (Widget,    widget),
   (XtPointer, closure),
   (XtPointer, callData) )
{
    fprintf(stderr, "Destroyed!\n");
    exit(0);
} /* Destroyed */


static void 
Quit  _P3( 
   (Widget,    widget),
   (XtPointer, closure),
   (XtPointer, callData) )
{
   t3d_end_space();
   XtDestroyWidget((Widget)closure);
   exit(0);
} /* Quit */


static void 
Newgame  _P3( 
   (Widget,    widget),
   (XtPointer, closure),
   (XtPointer, callData) )
{
   if (psrs != PSRS_PAUSE) {
      XtVaSetValues( pause_restart, XtNlabel, XtNpause, NULL);
      if (!psrs != PSRS_NONE) {
         XtMapWidget( pause_restart);
      }
      psrs = PSRS_PAUSE;
   }
   t3d_new_game();
   t3d_stop_timer();
   t3d_start_timer( NULL, NULL);
   XWarpPointer( XtDisplay(game), None /* src_w */, XtWindow(game) /* dest_w */,
                 0, 0, 0, 0, /* src_ <x, y, width, height> */
                 1, 1 /* dest_ <x, y> -  Just to be inside */);
} /* Newgame */


static void
endgame()
{
   DASSERT( psrs != PSRS_NONE);
   psrs = PSRS_NONE;
   XtUnmapWidget( pause_restart);
   t3d_stop_timer();
} /* endgame */


static void 
Pause_Restart  _P3( 
   (Widget,    widget),
   (XtPointer, closure),
   (XtPointer, callData) )
{
   DASSERT( psrs != PSRS_NONE);
   if (psrs == PSRS_PAUSE) {
      psrs = PSRS_RESTART;
      XtVaSetValues( pause_restart, XtNlabel, XtNrestart, NULL);

      t3d_stop_timer();
   } else {
      DASSERT( psrs == PSRS_RESTART);
      psrs = PSRS_PAUSE;
      XtVaSetValues( pause_restart, XtNlabel, XtNpause, NULL);

      t3d_start_timer( NULL, NULL);
      XWarpPointer( XtDisplay(game), None /* src_w */, 
                    XtWindow(game) /* dest_w */,
                    0, 0, 0, 0, /* src_ <x, y, width, height> */
                    1, 1 /* dest_ <x, y> -  Just to be inside */);
   }
} /* Pause_Restart */


static void 
redraw  _P4( 
   (Widget,    widget),
   (XtPointer, closure),
   (XEvent *,  event),
   (Boolean *, continue_to_dispatch)
)
{
   if ( event->xexpose.count == 0 ) {/* last Expose event */
      t3d_space_redraw();
   }
} /* Newgame */


int 
main  _P2( (int, argc),  (char **, argv))
{

   static char   *T3Dname = "T3D";
   static String fallback_resources[] = { 
                    "*scrollbar*length: 100",
                    "*game*width:       400",
                    "*game*height:      400",
                    "*highscorefile*width:   480",
                    "*highscorefile*height:  150",
                    NULL
                 };
   String trans = 
"<Key>L:       moveright()  \n\
 <Key>J:       moveleft()   \n\
 <Key>I:       moveup()     \n\
 <Key>K:       movedown()   \n\
 <Key>A:       rotpx()      \n\
 <Key>S:       rotpy()      \n\
 <Key>D:       rotpz()      \n\
 <Key>Q:       rotnx()      \n\
 <Key>W:       rotny()      \n\
 <Key>E:       rotnz()      \n\
 <Key>space:   dropdown()";

   static XtActionsRec  actions[] = {
                           {"moveright", t3d_act_right},
                           {"moveleft",  t3d_act_left},
                           {"moveup",    t3d_act_up},
                           {"movedown",  t3d_act_down},
                           {"rotpx",     t3d_act_rot_px},
                           {"rotpy",     t3d_act_rot_py},
                           {"rotpz",     t3d_act_rot_pz},
                           {"rotnx",     t3d_act_rot_nx},
                           {"rotny",     t3d_act_rot_ny},
                           {"rotnz",     t3d_act_rot_nz},
                           {"dropdown",  t3d_act_drop_down}
                        };
#  include "icon.h"
   XtTranslations   compiled_trans;
   Widget        toplevel, 
                 box,
                 menu,
                 quit,
                 newgame,
                 view;
   XtAppContext  app_con;
   Pixmap        t3d_icon;

   (void)fprintf(stderr, "%s - version 0.0. Copyright 1992 Yotam Medini\n",
                         T3Dname);
   toplevel = XtVaAppInitialize(&app_con, "Xds", NULL, 0,
                              &argc, argv, fallback_resources, NULL, 0);

   XtAddCallback( toplevel, XtNdestroyCallback, Destroyed, NULL );

   compiled_trans = XtParseTranslationTable(trans);
   DPR0((stderr, "compiled_trans=%x\n", compiled_trans));

   box = XtVaCreateManagedWidget( "box", boxWidgetClass, toplevel, 
                                  XtNorientation, XtorientHorizontal,
                                  NULL);

   menu = XtCreateManagedWidget( "menu", boxWidgetClass, box, NULL , 0);

   game = XtCreateManagedWidget( "game", coreWidgetClass, box, NULL, 0 );
   XtAddEventHandler( game, ExposureMask, FALSE, redraw, 0 );
   XtAugmentTranslations( game, compiled_trans );
    
   XtAppAddActions( app_con, actions, XtNumber(actions));
   newgame = XtVaCreateManagedWidget( "newgame", commandWidgetClass, menu, 
                                      XtNlabel, "Newgame",
                                      NULL);
   pause_restart = XtCreateManagedWidget( "pausrt", commandWidgetClass, menu, 
                                          NULL, 0);
   quit = XtVaCreateManagedWidget( "quit", commandWidgetClass, menu, 
                                   XtNlabel, "Quit", 
                                   NULL);
   t3d_score_init( app_con, toplevel, menu);
   view = XtVaCreateManagedWidget( "view", commandWidgetClass, menu, 
                                   XtNlabel, "View",
                                   NULL);
   t3d_endgame_callback( endgame);

   XtAddCallback( quit, XtNcallback, Quit, (XtPointer) toplevel);
   XtAddCallback( newgame, XtNcallback, Newgame, (XtPointer) toplevel);
   XtAddCallback( pause_restart, XtNcallback, Pause_Restart, 
                  (XtPointer) toplevel);
   XtRealizeWidget(toplevel);

   XtMapWidget(toplevel);

   psrs = PSRS_NONE;
   XtUnmapWidget( pause_restart);

   t3d_colorid_init( toplevel );
   t3d_space_init( toplevel, game, T3Dname);
   t3d_xview_init( toplevel, view);
   t3d_space_post();

   t3d_start_timer( app_con, toplevel);
   t3d_stop_timer();

   t3d_icon =  XCreateBitmapFromData( XtDisplay( box), XtWindow( box),
				      (char *)icon_bits, 
                                      icon_width, icon_height);
   XtVaSetValues( toplevel, XtNiconPixmap ,t3d_icon, NULL);

   XtAppMainLoop( app_con );
} /* main */
