#define DEBUG 0
#if 0

  Orientations -  matrices & rotation table initialization.

#endif

#include <memory.h>
#include "stdefun.h"
#include "orient.h"
#include "debug.h"


/*  For temporary 3 (6*6 = 36) tables with  0..T3D_N_ORIENT-1 inidices,
 *  1st index, for which axe is  n o t  in the double axis/dir-key
 *
 *  So, given  axe, let axe1, axe2 be modulo-3 successors, then
 *    QRI_XY [axe] [i1][d1] [i2][d2] = the orientation index, 
 *  such that axe_j goes to axe - i_j in (-1)^d_j  direction (j=1,2).
 */
typedef  int  QRI_XY [3] [3][2] [3][2]; 

/* calculate all 24 possible combination of orthonormal unit vectors */
static void
init_quad_rots  _P2(
   (T3D_MATQR *, orient),
   (QRI_XY,     qri_xy))
{
   int          ix0,        /* 1st axe */ 
                id0,        /* ist dir */
                ix1o, ix1,  /* 2nd axe */
                id1,        /* 2nd dir */
                ix2,        /* 3rd axe */
                id2,        /* 3rd dir */
                n = 0;      /* quad_rot index */
   T3D_QUAD_ROT  *pqd;

#if DEBUG /* Dirty! */
{int i, *ia = (int *)qri_xy;
 for (i=0; i< 3* 3*2 * 3*2; i++) ia[i] = -1992;
}
#endif /* DEBUG

   /* 3*2*2*2 = 24 in 4-for-loops */
   for (ix0 = 0;  ix0 < 3;  ix0++) {  
      for (id0 = 0;  id0 < 2;  id0++) {

         for (ix1o = 1;  ix1o <= 2;  ix1o++) {
            ix1 = (ix0 + ix1o) % 3;
            for (id1 = 0;  id1 < 2;  id1++) {

               pqd = &orient[n].qr;

               pqd->oxyz[0].axe = ix0;
               pqd->oxyz[0].dir = id0;
               pqd->oxyz[1].axe = ix1;
               pqd->oxyz[1].dir = id1;

               /* the 3rd is enforced! */         
               pqd->oxyz[2].axe = ix2 = (0+1+2) - (ix0 + ix1);
               pqd->oxyz[2].dir = id2 = (ix0 + (ix2 < ix1) + id0 + id1) % 2;

               /*  Any 2 axes uniquly define an orientation, 
                *  so we make any two a key!
                */
               qri_xy [0] [ix1][id1] [ix2][id2] = n;
               qri_xy [1] [ix2][id2] [ix0][id0] = n;
               qri_xy [2] [ix0][id0] [ix1][id1] = n;
               n++;
            }
         }
      }
   }

   DASSERT(n == T3D_N_ORIENT);
#if DEBUG
{int k;
 for (k = 0;  k < 3;  k++) {
 printf("\n\nTable keys[Non- %d]\n", k);
  for (ix0 = 0;  ix0 < 3;  ix0++) {  
   for (id0 = 0;  id0 < 2;  id0++) {
    for (ix1 = 0;  ix1 < 3;  ix1++) {
     for (id1 = 0;  id1 < 2;  id1++) {
      printf("[%d][%d] [%d][%d] = %d\n",      ix0, id0,  ix1, id1,
                                   qri_xy[k] [ix0][id0] [ix1][id1]);
     }
    }
   }
  }
 }
}
#endif /* DEBUG */
} /* init_quad_rots */



/* Establish map : (orientation, quad-rotation) -> orientation */
static void
init_rotquad_table  _P3(
   (T3D_ROTQUAD_TABLE *, rqo),
   (T3D_MATQR *,         orient),
   (QRI_XY,             qri_xy))
{
   int         i, tfx;
   T3D_ROTQUAD  rq;

   DASSERT(T3D_N_ROTQUAD == 6);
   for (rq = T3D_ROTQUAD_PX;  rq < T3D_N_ROTQUAD;  rq++) {
      tfx = rq / 2;  /* The fixed axe of the rotation */

      /*  For each orientation, we will see, what happens with:
       *  -  tfx  (fx, fd)
       *  -  axe1       (sx, sd)  after the quad rotation (taken from axe2).
       */
      for (i = 0;  i < T3D_N_ORIENT;  i++) {
         int  oi, fx, fd, sx, tsx, sd, kx;

         /* Who goes totally to the  fixed axis */
         fx = 3;
         do {
            fx--;  DASSERT(fx >= 0);
         } while (orient[i].qr.oxyz[fx].axe != tfx);
         fd = orient[i].qr.oxyz[fx].dir; 

         sx = (fx + 1) % 3;  /* Rotating axis */
         kx = (sx + 1) % 3;  /* Other rotating, let it be a key axis */
         /*  The rotation sign Sr = (rq % 2), if positive (Sr=0) then get the
          *  direction of kx,  if negative (Sr=1) then the opposite.
          */
         tsx = orient[i].qr.oxyz[kx].axe;
         sd = (orient[i].qr.oxyz[kx].dir + rq) % 2;

         /* Since the key is the (fx, sx) axe-pair, we take kx as the key ! */
         DASSERT( (fx+1)%3 == sx  &&  (sx+1)%3 == kx  &&  (kx+1)%3 == fx );
         rqo->orient[i][rq] = oi = qri_xy [kx] [tfx][fd] [tsx][sd];
         DASSERT(0 <= oi && oi < T3D_N_ORIENT);
      }
   }

#if DEBUG
   printf("\n\nRotation Orientation index - table\n");
   printf("Orig-orient |  +X  -X  +Y  -Y  +Z  -Z\n");  
   printf("======================================");
   for (i = 0;  i < T3D_N_ORIENT;  i++) {
      printf("\n   %2d       |", i);
      for (rq = T3D_ROTQUAD_PX;  rq < T3D_N_ROTQUAD;  rq++) {
         /* Test if rq^4 = Identity */
         int ri, oi;
         for (ri = 0, oi = i; ri < 4;  ri++)  oi = rqo->orient[oi][rq];
         if (oi != i) printf(" Error! (rq=%d)^4(i=%d) != (oi=%d)\n",
                                      rq,      i,        oi);
         printf("  %2d", rqo->orient[i][rq]);
         
      }
   }
   printf("\n");
#endif
} /* init_rotquad_table */



static void 
mat3_of_unitvector  _P2(
   (T3D_MAT3 *,     pmat),
   (T3D_QUAD_ROT *, quad_rot))
{
   static T3D_MAT3  zmat = {{
                             {0., 0., 0.},
                             {0., 0., 0.},
                             {0., 0., 0.}  
                          }};
   static float    sign[2] = {1., -1.};
   int  i;

   (void)memcpy(pmat, &zmat, sizeof(T3D_MAT3));  /* Zeroing */

   for (i = 0;  i < 3;  i++) {
      pmat->a [i] [ quad_rot->oxyz[i].axe ] = sign[ quad_rot->oxyz[i].dir ];
   }

} /* mat3_of_unitvector */
                 


void 
t3d_set_orient_table  _P2(
   (T3D_MATQR *,         orient),
   (T3D_ROTQUAD_TABLE *, rqo) )
{
   int          i;
   QRI_XY       qri_xy;

   init_quad_rots( orient, qri_xy);

   init_rotquad_table( rqo, orient, qri_xy);

#ifdef NEED_ORMAT
   for (i = 0;  i < T3D_N_ORIENT;  i++) {
      mat3_of_unitvector(&orient[i].mat, &orient[i].qr);

#if DEBUG
{        int j, k;
         printf("\n quad_rot[%d] = (", i);
         for (j = 0;  j < 3;  j++) {
            printf("%ce%1d ", "+-"[orient[i].qr.oxyz[j].dir],
                                   orient[i].qr.oxyz[j].axe);
         }
         printf(")\n\t", i);
         for (j = 0;  j < 3;  j++) {
            for (k = 0;  k < 3;  k++) {
               printf("%3g  ", orient[i].mat.a[j][k]);
            }
            printf("\n\t");
         }
}
#endif /* DEBUG */

   }
#endif /* NEED_ORMAT */
} /* t3d_set_orient_table */                    
