#include <stdio.h>
#include <string.h>
#ifdef unix
#  include <pwd.h>
#  include <time.h>
#endif
#include <sys/types.h>
#include <sys/times.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Box.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Paned.h>
#include "stdefun.h"
#include "score.h"
#include "debug.h"

enum {
   fn_size = 120,
   score_st_size = 25
};


typedef struct {
   Widget  popper,
           popped;
} POPAIR;

typedef struct {
   int   score;
   char  name[score_st_size];
   char  when[score_st_size];
} SCOREC;

static int      score;
static char     high_score_fn[ fn_size + 1];
static Widget   currscore, scorefile;
static float    coef[T3D_SCOREVENT_N][3];

static void
show_score()
{
   char message[80];  /* must be enough! */
   sprintf( message, "%d", score);
   XtVaSetValues( currscore, XtNlabel, message, NULL);
} /* show_score */


static void
Score_reload  _P3(
   (Widget,    widget),
   (XtPointer, closure),
   (XtPointer, callData) )
{
   XtVaSetValues( (Widget)closure, XtNstring, high_score_fn, NULL);
} /* Score_reload */


static void
ScoreDone  _P3(
   (Widget,    widget),
   (XtPointer, closure),
   (XtPointer, callData) )
{
   POPAIR   *p = (POPAIR *)closure;
   XtPopdown( p->popped);
   XtSetSensitive( p->popper, 1);
} /* ScoreDone */


#ifdef unix


static char *
me()
{
   static int  virgin = 1;
   static char rc[score_st_size];
   int         nn, ng;
   if (virgin) {
      struct passwd  *pwd = getpwuid( getuid());
      virgin = 0;
      (void)strncpy( rc,
                     (strlen(pwd->pw_name) >= strlen(pwd->pw_gecos)
                        ?  pwd->pw_name
                        :  pwd->pw_gecos),
                     score_st_size);
   }
   return( rc );
} /* me */


static char *
now()
{
   time_t   t;
   t = time(NULL);
   return( t >= 0 ? ctime(&t) : "Unknow-time");
} /* now */
#endif /* unix */


static void
spacedel  _P2(
   (char *, s),
   (char,   del)
)
{
   for (; *s; s++) {
      if (*s == del) {
         *s = ' ';
      }
   }
} /* spacedel */


static void
print_score  _P2(
   (FILE *,   fp),
   (SCOREC *, psc)
)
{
   psc->name[ score_st_size-1 ] =
   psc->when[ score_st_size-1 ] = '\0';

   spacedel( psc->name, '|');
   spacedel( psc->when, '|');

   DASSERT( 25 == score_st_size);
   (void)fprintf( fp, "%10d | %-25s | %-25s\n",
                      psc->score, psc->name, psc->when);
} /* print_score */


void
t3d_score_init  _P3(
   (XtAppContext, app_con),
   (Widget,       toplevel),
   (Widget,       father)
)
{
   static XtResource hsfn_res = {
                        "highScoreFilename",
                        "HighScoreFilename",
                        XtRString,
                        sizeof(char *),
                        0,
                        XtRImmediate,
                        "t3d.score"
                     };
   static XtResource poly_res[] = { /* According to T3D_SCOREVENT enumeration */

                        {  "newPiece0", "NewPiece0", XtRFloat, sizeof(float),
                           0*sizeof(float), XtRString, "5."
                        },
                        {  "newPiece1", "NewPiece1", XtRFloat, sizeof(float),
                           1*sizeof(float), XtRString, ".3"
                        },
                        {  "newPiece2", "NewPiece2", XtRFloat, sizeof(float),
                           2*sizeof(float), XtRString, ".1"
                        },


                        {  "drop0", "Drop0", XtRFloat, sizeof(float),
                           3*sizeof(float), XtRString, ".1"
                        },
                        {  "drop1", "Drop1", XtRFloat, sizeof(float),
                           4*sizeof(float), XtRString, "1."
                        },
                        {  "drop2", "Drop2", XtRFloat, sizeof(float),
                           5*sizeof(float), XtRString, ".01"
                        },


                        { "1level0", "1Level0", XtRFloat, sizeof(float),
                           6*sizeof(float), XtRString, "10."
                        },
                        {  "1level1", "1Level1", XtRFloat, sizeof(float),
                           7*sizeof(float), XtRString, ".01"
                        },
                        {  "1level2", "1Level2", XtRFloat, sizeof(float),
                           8*sizeof(float), XtRString, "0."
                        },

                        { "2level0", "2Level0", XtRFloat, sizeof(float),
                           9*sizeof(float), XtRString, "21."
                        },
                        {  "2level1", "2Level1", XtRFloat, sizeof(float),
                           10*sizeof(float), XtRString, ".01"
                        },
                        {  "2level2", "2Level2", XtRFloat, sizeof(float),
                           11*sizeof(float), XtRString, "0."
                        },

                        { "3level0", "3Level0", XtRFloat, sizeof(float),
                           12*sizeof(float), XtRString, "31."
                        },
                        {  "3level1", "3Level1", XtRFloat, sizeof(float),
                           13*sizeof(float), XtRString, ".1"
                        },
                        {  "3level2", "3Level2", XtRFloat, sizeof(float),
                           14*sizeof(float), XtRString, "0."
                        },

                        { "4level0", "4Level0", XtRFloat, sizeof(float),
                           15*sizeof(float), XtRString, "41."
                        },
                        {  "4level1", "4Level1", XtRFloat, sizeof(float),
                           16*sizeof(float), XtRString, ".1"
                        },
                        {  "4level2", "4Level2", XtRFloat, sizeof(float),
                           17*sizeof(float), XtRString, "0.1"
                        }
                     };
   static POPAIR     popair;
   char              *hfn;
   Widget            score, highscores, popup, scorebox,
                     scoreDone, reload;

   XtGetApplicationResources( father, &hfn, &hsfn_res, 1, NULL, 0);
   (void)strncpy( high_score_fn, hfn, fn_size - 1);

   DASSERT( XtNumber(poly_res) == 3 * XtNumber(coef) );
   XtGetApplicationResources( father, coef, poly_res, XtNumber(poly_res),
                              NULL, 0);

   score = XtCreateManagedWidget( "score", boxWidgetClass, father, NULL, 0);
   (void)XtVaCreateManagedWidget( "label", labelWidgetClass, score,
                                  XtNlabel, "Score", NULL);
   currscore = XtVaCreateManagedWidget( "current", labelWidgetClass, score,
                                        XtNlabel, "0", NULL );
   highscores = XtVaCreateManagedWidget( "highs", commandWidgetClass, score, 
                                         XtNlabel, "Highs",
                                         NULL);

   popup = XtCreatePopupShell( "High Scores", transientShellWidgetClass,
                               toplevel, NULL, 0);
   scorebox = XtCreateManagedWidget( "scorebox", panedWidgetClass, popup,
                                     NULL, 0);
   scoreDone = XtCreateManagedWidget( "Close", commandWidgetClass, scorebox,
                                      NULL, 0);
   reload = XtCreateManagedWidget( "Reload", commandWidgetClass, scorebox,
                                   NULL, 0);
   scorefile = XtVaCreateManagedWidget( "highscorefile", asciiTextWidgetClass,
                                        scorebox,
                                        XtNeditType, XawtextRead,
                                        XtNtype, XawAsciiFile,
                                        XtNstring, high_score_fn,
                                        NULL);
   XtAddCallback( highscores, XtNcallback, XtCallbackNone, popup);
   XtAddCallback( highscores, XtNcallback, Score_reload, scorefile);

   popair.popper = highscores;
   popair.popped = popup;
   XtAddCallback( scoreDone, XtNcallback, ScoreDone, &popair);

   XtAddCallback( reload, XtNcallback, Score_reload, scorefile);
} /* t3d_score_init */


void
t3d_score_set_value  _P1(
   (int,  v)
)
{
   score = v;
   show_score();
} /* t3d_score_set_value */


void
t3d_score_add_value  _P1(
   (int,  v)
)
{
   score += v;
   show_score();
} /* t3d_score_add_value */


int
t3d_score_event_value  _P2(
   (T3D_SCOREVENT,  e),
   (int,           n)
)
{
   float  x = n,
          *a = coef[e],
          v = (a[2]*x + a[1])*x + a[0];
   int    rc = (int)v;
   return(rc);
} /* t3d_score_event_value */


void
t3d_score_add_event_value  _P2(
   (T3D_SCOREVENT,  e),
   (int,           n)
)
{
   t3d_score_add_value( t3d_score_event_value( e, n));
} /* t3d_score_add_event_value */


void
t3d_score_record()
{
   int     i, n, ok;
   FILE    *fp;
   SCOREC  top10[10], nowme;

   fp = fopen(high_score_fn, "r");
   if (!fp) {
      fprintf(stderr, "Opening new High-Score file %s\n", high_score_fn);
      n = 0;
   } else {
      for (i = 0, ok = 1;  i < 10 && ok;  i++) {
         DASSERT( 25 == score_st_size);
#if 0
         ok = fscanf( fp, "%d : %25s : %25s\n", &top10[i].score,
                          top10[i].name, top10[i].when) == 3;
#endif
         ok = fscanf( fp, "%d |", &top10[i].score) == 1;
         ok = ok && fscanf( fp, " %25[^|] |", top10[i].name) == 1;
         ok = ok && fscanf( fp, " %25[^\n]\n", top10[i].when) == 1;
         /* Decreasing */
         DASSERT(!ok || i == 0 || top10[i].score <= top10[i-1].score);
      }
      fclose(fp);
      n = (ok ? i : i - 1);
   }

   if (n < 10 || score > top10[10-1].score) { /* Enter Top-10 !! */


         nowme.score = score;
#ifdef unix
         (void)strncpy( nowme.name, me(), score_st_size);
         (void)strncpy( nowme.when, now(), score_st_size);
#else
         (void)strcpy( nowme.name, "Non-unix-WHO");
         (void)strcpy( nowme.when, "Non-unix-WHEN");
#endif /* unix */


      fp = fopen(high_score_fn, "w");
      if (!fp) {
         (void)fprintf(stderr, "Could not open %s for recording score\n",
                         high_score_fn);
      } else {
         for (i = 0;  i < n && top10[i].score > score;  i++) {
             print_score( fp, &top10[i]);
         }
         print_score( fp, &nowme);
         if (n >= 10)  n--;  /* previous last is kicked out */
         for (;  i < n;  i++) {
             print_score( fp, &top10[i]);
         }
         fclose(fp);
         XtVaSetValues( scorefile, XtNstring, high_score_fn, NULL);
      }
   }
} /* t3d_score_add_value */


