#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include "stdefun.h"
#include "timer.h"
#include "game.h"
#include "debug.h"


static XtIntervalId         timer = NULL;


static void
clock_call _P2(
   (XtPointer,  client_data),
   (XtIntervalId *,  timer)
)
{
   DASSERT(timer != NULL);
   t3d_fall();
   t3d_start_timer( NULL, NULL);
} /* clock_call */


void 
t3d_start_timer _P2(
   (XtAppContext, app_con),
   (Widget,       w)
)
{
   static XtAppContext  the_app_con = NULL;
   static int           a, b, c, m;
   int                  dt;

   DASSERT((app_con == NULL) != (the_app_con == NULL));
   DASSERT((app_con == NULL) == (w == NULL));

   if (!the_app_con) {
      /* first time, get millisecPerFall<> from resource */
      typedef struct {
         int  begin,  end, /* millisec / fall rates */
              med_fall;    /* # fall, to get the median */
      } MPS, *MPSP;
      MPS  mps;
      static XtResource  period_res[] = {
                            {
                               "millisecPerFallBegin",
                               "MillisecPerFallBegin",
                               XtRInt,
                               sizeof(int),
                               XtOffset(MPSP, begin),
                               XtRImmediate,
                               (XtPointer)1500
                            },
                            {
                               "millisecPerFallEnd",
                               "MillisecPerFallEnd",
                               XtRInt,
                               sizeof(int),
                               XtOffset(MPSP, end),
                               XtRImmediate,
                               (XtPointer)500
                            },
                            {
                               "millisecPerFallMedWhen",
                               "MillisecPerFallMedWhen",
                               XtRInt,
                               sizeof(int),
                               XtOffset(MPSP, med_fall),
                               XtRImmediate,
                               (XtPointer)50
                            },
                         };
      the_app_con = app_con;
      XtGetApplicationResources( w, &mps, period_res, XtNumber(period_res), 
                                 NULL, 0);
      a = mps.med_fall * (mps.begin - mps.end);
      b = mps.med_fall;  if (b == 0) b = 1;
      c = mps.end;
   }
   m = t3d_game_npiece() - 1;
   dt = a / (b + m) + c;
   timer = XtAppAddTimeOut( the_app_con, dt, clock_call, NULL);
   t3d_game_set_active(1);
} /* t3d_start_timer */


void 
t3d_stop_timer()
{
   if (timer) {
      XtRemoveTimeOut( timer);
      timer = NULL;
      t3d_game_set_active(0);
   }
} /* t3d_stop_timer */
